Attribute VB_Name = "serapi"
'Single example serial port handle and overlapped structure

    Public hPort As Long
    Public olWritePort As OVERLAPPED
    Public olWritePending As Boolean
    Public olWriteBuffer As String
    Public olReadPort As OVERLAPPED
    
' Bare minimum Win32 API and type declares for serial port access
'
Public Const INVALID_HANDLE_VALUE = -1  'indicates CreateFile() failure
Public Const MAXDWORD = &HFFFF          'needed to SetCommTimeouts()
'
Public Const ERROR_IO_INCOMPLETE = 996& 'Expected values of GetLastError()
Public Const ERROR_IO_PENDING = 997&    'for the overlapped transmit I/O
'
' This is how Microsoft sets up File_Generic_* which we need for CreateFile()
' Essentially a very complicated way to get &h120089 and &h120116
'
Public Const READ_CONTROL = &H20000
Public Const STANDARD_RIGHTS_READ = (READ_CONTROL)
Public Const STANDARD_RIGHTS_WRITE = (READ_CONTROL)
Public Const FILE_READ_DATA = (&H1)
Public Const FILE_READ_ATTRIBUTES = (&H80)
Public Const FILE_READ_EA = (&H8)
Public Const FILE_WRITE_DATA = (&H2)
Public Const FILE_WRITE_ATTRIBUTES = (&H100)
Public Const FILE_WRITE_EA = (&H10)
Public Const FILE_APPEND_DATA = (&H4)
Public Const SYNCHRONIZE = &H100000
'
Public Const FILE_GENERIC_READ = _
    (STANDARD_RIGHTS_READ Or FILE_READ_DATA Or FILE_READ_ATTRIBUTES Or _
    FILE_READ_EA Or SYNCHRONIZE)
'
Public Const FILE_GENERIC_WRITE = _
    (STANDARD_RIGHTS_WRITE Or FILE_WRITE_DATA Or FILE_WRITE_ATTRIBUTES Or _
    FILE_WRITE_EA Or FILE_APPEND_DATA Or SYNCHRONIZE)
'
Public Const FILE_FLAG_OVERLAPPED = &H40000000  'Other standard CreateFile()
Public Const OPEN_EXISTING = 3                  'control values
'
' These types are required, set up, and used by some of the
' API functions, but this code generally doesn't reference
' their individual fields directly.

' Required by CreateFile() but not really used in this application
'
Type SECURITY_ATTRIBUTES
        nLength As Long
        lpSecurityDescriptor As Long
        bInheritHandle As Long
End Type

' You need one to allocate one of these for each overlapped operation,
' and the OS will use them to keep track of which operation is which.
'
Type OVERLAPPED
        Internal As Long
        InternalHigh As Long
        offset As Long
        OffsetHigh As Long
        hEvent As Long
End Type

' Used to pass setup parameters to SetCommState(), we won't reference
' it directly but use BuildCommDCB() to create it so our code won't
' break when Microsoft changes the specification uh-gain
'
Type DCB
        DCBlength As Long
        BaudRate As Long
        fBitFields As Long
        wReserved As Integer
        XonLim As Integer
        XoffLim As Integer
        ByteSize As Byte
        Parity As Byte
        StopBits As Byte
        XonChar As Byte
        XoffChar As Byte
        ErrorChar As Byte
        EofChar As Byte
        EvtChar As Byte
        wReserved1 As Integer
End Type
'
' We actually set and read some of the fields in these types.
'
' This type is used to force ReadFile() to return with BytesRead=0
' when there is no data available instead of blocking until the
' specified number of bytes arrive.
'
Type COMMTIMEOUTS
        ReadIntervalTimeout As Long
        ReadTotalTimeoutMultiplier As Long
        ReadTotalTimeoutConstant As Long
        WriteTotalTimeoutMultiplier As Long
        WriteTotalTimeoutConstant As Long
End Type
'
' This type is used by ClearCommError() (!?!?!?) to return the
' bytes waiting unprocessed in input and output buffers.
'
Type COMSTAT
        fBitFields As Long
        cbInQue As Long
        cbOutQue As Long
End Type
'
Declare Function CreateFile Lib "kernel32" Alias "CreateFileA" _
    (ByVal lpFileName As String, ByVal dwDesiredAccess As Long, _
    ByVal dwShareMode As Long, lpSecurityAttributes As SECURITY_ATTRIBUTES, _
    ByVal dwCreationDisposition As Long, ByVal dwFlagsAndAttributes As Long, _
    ByVal hTemplateFile As Long) As Long
'
Declare Function ReadFile Lib "kernel32" _
    (ByVal hFile As Long, lpBuffer As Any, _
    ByVal nNumberOfBytesToRead As Long, lpNumberOfBytesRead As Long, _
    lpOverlapped As OVERLAPPED) As Long
'
Declare Function WriteFile Lib "kernel32" _
    (ByVal hFile As Long, lpBuffer As Any, _
    ByVal nNumberOfBytesToWrite As Long, lpNumberOfBytesWritten As Long, _
    lpOverlapped As OVERLAPPED) As Long
'
Declare Function CloseHandle Lib "kernel32" _
    (ByVal hObject As Long) As Long
'
Declare Function BuildCommDCB Lib "kernel32" Alias "BuildCommDCBA" _
    (ByVal lpDef As String, lpDCB As DCB) As Long
'
Declare Function SetCommState Lib "kernel32" _
    (ByVal hCommDev As Long, lpDCB As DCB) As Long
'
Declare Function GetCommState Lib "kernel32" _
    (ByVal hCommDev As Long, lpDCB As DCB) As Long
'
Declare Function SetCommTimeouts Lib "kernel32" _
    (ByVal hFile As Long, lpCommTimeouts As COMMTIMEOUTS) As Long
'
Declare Function GetCommTimeouts Lib "kernel32.dll" _
    (ByVal hFile As Long, lpCommTimeouts As COMMTIMEOUTS) As Long

Declare Function ClearCommError Lib "kernel32" _
    (ByVal hFile As Long, lpErrors As Long, lpStat As COMSTAT) As Long
'
Declare Function GetLastError Lib "kernel32" () As Long
'
Declare Function GetOverlappedResult Lib "kernel32" _
    (ByVal hFile As Long, lpOverlapped As OVERLAPPED, _
    lpNumberOfBytesTransferred As Long, ByVal bWait As Long) As Long

Function OpenPort(PName As String, PSettings As String) As String
    '
    'This function returns a null string "" if it succeeds,
    'otherwise it returns an error which can be MsgBoxed or
    'ignored as appropriate
    '
    Dim RVal As Long, PortStr As String
    Dim PSA As SECURITY_ATTRIBUTES, PortOL As OVERLAPPED
    Dim Timeouts As COMMTIMEOUTS, ADCB As DCB
    '
    'First we must get a handle for the port by name.  NB under
    'WinNT etc. ports do not have to be named COMnn, which is
    'why the port name is a string.
    '
    hPort = CreateFile("\\.\" & PName, _
        FILE_GENERIC_WRITE Or FILE_GENERIC_READ, _
        0, PSA, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, 0)
    If hPort = INVALID_HANDLE_VALUE Then
        OpenPort = "Can't Open Port " + PName + "."
        Exit Function
    End If
    '
    RVal = GetCommState(hPort, ADCB)
    If RVal = 0 Then
        Call CloseHandle(hPort)
        OpenPort = "Couldn't read port " + PName + " settings."
        Exit Function
    End If
    '
    'And we set the settings as specified ("9600,n,8,1" or whatever)
    'Fill in the DCB...
    '
    RVal = BuildCommDCB(PSettings, ADCB)
    If RVal = 0 Then
        Call CloseHandle(hPort)
        OpenPort = "Port " & PName & " DCB Build Fail: " & PSettings
        Exit Function
    End If
    '
    '...and use the DCB to set up the port.
    '
    RVal = SetCommState(hPort, ADCB)
    If RVal = 0 Then
        Call CloseHandle(hPort)
        OpenPort = "Port " & PName & " settings didn't take: " & PSettings
        Exit Function
    End If
    '
    'Now we set the timeouts so that ReadFile() will return
    'immediately instead of blocking if there is no data.
    '
    RVal = GetCommTimeouts(hPort, Timeouts)
    If RVal = 0 Then
        Call CloseHandle(hPort)
        OpenPort = "Couldn't read port " + PName + " tiemouts."
        Exit Function
    End If
    Timeouts.ReadIntervalTimeout = MAXDWORD
    Timeouts.ReadTotalTimeoutMultiplier = 0
    Timeouts.ReadTotalTimeoutConstant = 0
    Timeouts.WriteTotalTimeoutMultiplier = 0
    Timeouts.WriteTotalTimeoutConstant = 0
    RVal = SetCommTimeouts(hPort, Timeouts)
    If RVal = 0 Then
        Call CloseHandle(hPort)
        OpenPort = "Couldn't set port " + PName + " timeouts."
        Exit Function
    End If
    '
    OpenPort = ""    'It worked!
    '
End Function

Sub ClosePort()
    'Close the port.
    Call CloseHandle(hPort)
    'Gee, that was hard.
End Sub

Function RXport() As String
    '
    'Receive all available characters from the serial port.
    '
    Const MsgChunkSize As Integer = 255
    'leave room for null termination...
    Dim AByte(MsgChunkSize + 1) As Byte
    Dim RVal As Long, BytesRead As Long
    Dim MsgChunk As String, Msg As String
    Msg = ""
    Do
        RVal = ReadFile(hPort, AByte(0), MsgChunkSize, BytesRead, olReadPort)
        If RVal = 0 Then
            BytesRead = 0      'Terminate DO loop
        ElseIf BytesRead > 0 Then
            MsgChunk = StrConv(AByte, vbUnicode)
            If Len(MsgChunk) > BytesRead Then MsgChunk = Left$(MsgChunk, BytesRead)
            Msg = Msg & MsgChunk
        End If
        'If the chunk was exactly the size requested, there may be more
    Loop While BytesRead = MsgChunkSize
    RXport = Msg
    '
End Function

Sub TXport(Msg As String)
    '
    'Send the message to the serial port.  This fails transparently if
    'an overlapped transmit is still in progress.  Calling code should
    'check TXportBusy before calling and maintain its own queue if
    'this is not appropriate.
    '
    'If you want to tack successive Msg's onto the outbound buffer,
    'there are two approaches.  One is to call another function to
    'append a homemade buffer and poll TXportBusy() to see when it
    'is safe to shovel more bytes here.  This is completely safe
    'but requires a timer object and will introduce small delays of
    'up to the timer period.
    '
    'More efficient and more Windows-y would be to use the Windows API
    'callback to issue another WriteFile() when the previous Overlapped
    'function is complete.  Unfortunately, this is a complete waste of
    'time because the Windows API functions for doing it are broken, and
    'lose notifications if the main program thread is locked when they
    'occur.  (The MSCOMM control itself suffers from this, and must be
    'polled with a timer to work reliably.)
    '
    'Even faster and more politically correct under the Windows paradigm
    'would be to alloacte multiple olWritePort structures so that multiple
    'overlapped WriteFiles could be called at the same time.  This would,
    'however, introduce an issue of how to allocate these structures.
    '
    'At the high (Visual Basic) level, the obvious technique is to create
    'a dynamic array of type OVERLAPPED and use ReDim Preserve to adjust
    'its size as necessary.  However, given the way it is used at the API
    'level I am fairly sure it is not safe for olWritePort to move around,
    'which is why I have declared it static at the module level here.
    'You could use a static array, but you would have to decide how large
    'to make it at design time.  You would also have to have control
    'structures for allocating entries and reacting gracefully when the
    'array is full.
    '
    'Whatever complexities are added, this is the basic technique:
    '
    Dim RVal As Long, MsgLen As Long, BytesWritten As Long
    '
    If TXportBusy Then Exit Sub
    MsgLen = Len(Msg)
    RVal = WriteFile(hPort, ByVal Msg, MsgLen, BytesWritten, olWritePort)
    '
    'RVal will just about always return an error due to the overlapped operation.
    'we could detect and verify it as ERROR_IO_PENDING or ERROR_IO_INCOMPLETE
    'via GetLastError() if we weren't so lazy.
    '
    olWritePending = True
    '
End Sub

Function TXportBusy() As Boolean
    '
    'Check to see if an overlapped transmit is in progress, which
    'will cause another transmit to fail.
    '
    If olWritePending Then
        RVal = GetOverlappedResult(hPort, olWritePort, BytesWritten, False)
        '
        'If RVal returns an error, we could detect and verify it as
        'ERROR_IO_PENDING or ERROR_IO_INCOMPLETE via GetLastError()
        'if we weren't so lazy.
        '
        If RVal = 0 Then
            TXportBusy = True 'previous operation most likely incomplete
        Else
            olWritePending = False
            TXportBusy = False
        End If
    Else
        TXportBusy = False
    End If
End Function


